SystemOrganization addCategory: #Loom!

Announcement subclass: #LoomEvent
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Loom'!

LoomEvent subclass: #LoomSampledEvent
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Loom'!

LoomEvent subclass: #LoomSamplingEvent
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Loom'!

TestCase subclass: #LoomTest
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Loom'!

LoomTest subclass: #LoomEventTest
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Loom'!

!LoomEventTest methodsFor: 'tests' stamp: 'AlexandreBergel 9/11/2012 22:17'!
testSampledEvent
	| t profiler |
	profiler := LoomProfiler new.
	t := 0.
	profiler on: LoomSamplingEvent do: [ :event | t := t + 1 ].
	profiler on: LoomSampledEvent do: [ :event | t := t + 10 ].
	
	self assert: t = 0.
	profiler sampleExecution.
	self assert: t = 1.! !

LoomTest subclass: #LoomPlatformTest
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Loom'!

!LoomPlatformTest methodsFor: 'tests' stamp: 'AlexandreBergel 9/6/2012 09:53'!
testLoomTime

	| time |
	time := LoomTime currentTimeAsMilliseconds.
	(Delay forMilliseconds: 2) wait.
	self assert: (LoomTime currentTimeAsMilliseconds - time ) >= 2! !

!LoomPlatformTest methodsFor: 'tests' stamp: 'AlexandreBergel 9/6/2012 10:00'!
testProcessor

	"Just to make sure we have what we expect"
	self assert: (Processor class == ProcessorScheduler)! !

LoomTest subclass: #LoomProfilerTest
	instanceVariableNames: 'thread1 thread2 profiler thread3'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Loom'!

!LoomProfilerTest methodsFor: 'util' stamp: 'AlexandreBergel 9/5/2012 14:23'!
isScenarioOver
	(Array with: thread1 with: thread2 with: thread3) 
		do: [ :t | (t notNil and: [ t isTerminated not ]) ifTrue: [ ^ false ] ].
	^ true
! !

!LoomProfilerTest methodsFor: 'scenarios' stamp: 'AlexandreBergel 9/6/2012 10:19'!
scenario1

	thread1 := [ 100 timesRepeat: [ self wait ] ] fork.
	thread2 := [ 50 timesRepeat: [ self wait2 ] ] fork.! !

!LoomProfilerTest methodsFor: 'scenarios' stamp: 'AlexandreBergel 9/5/2012 14:27'!
scenario2

	thread1 := [ 200 timesRepeat: [ self wait ] ] fork.
	thread2 := [ 100 timesRepeat: [ self wait2 ] ] fork.
	thread3 := [ 50 timesRepeat: [ self wait3 ] ] fork.! !

!LoomProfilerTest methodsFor: 'running' stamp: 'AlexandreBergel 9/5/2012 10:51'!
setUp
	super setUp.
	profiler := LoomProfiler new.! !

!LoomProfilerTest methodsFor: 'running' stamp: 'AlexandreBergel 9/5/2012 10:54'!
tearDown
	profiler stopProfiling! !

!LoomProfilerTest methodsFor: 'tests' stamp: 'AlexandreBergel 9/6/2012 10:22'!
testBasic

	self scenario1.
	profiler profile: (Array with: thread1 with: thread2).
	self waitForScenarioEnd.
	profiler stopProfiling.
	
	self assert: profiler profilingTime > 10.
	
	"Thread1 consume twice more the CPU than Thead2. Actually, thread2 is sleeping twice much."

	self assert: (profiler consumptionTimeOf: thread1) > (profiler consumptionTimeOf: thread2).
	self assert: (profiler consumptionTimeOf: thread1) >= 0.2.
	self assert: (profiler consumptionTimeOf: thread2) <= 0.7.! !

!LoomProfilerTest methodsFor: 'tests' stamp: 'AlexandreBergel 9/6/2012 10:24'!
testBasic2

	self scenario2.
	profiler profile: (Array with: thread1 with: thread2 with: thread3).
	self waitForScenarioEnd.
	profiler stopProfiling.
	
	"Thread1 consume twice more the CPU than Thead2. Actually, thread2 is sleeping twice much."
	self assert: (profiler consumptionTimeOf: thread1) > (profiler consumptionTimeOf: thread2).
	self assert: (profiler consumptionTimeOf: thread1) > (profiler consumptionTimeOf: thread3).
	self assert: (profiler consumptionTimeOf: thread3) >  0.
	self assert: (profiler consumptionTimeOf: thread2) >  0! !

!LoomProfilerTest methodsFor: 'tests' stamp: 'AlexandreBergel 9/6/2012 08:41'!
testGetLoomThreadFrom
	| p ps |
	p := [ ] fork.

	profiler profile: (Array with: p).
	self assert: (profiler getThreadFrom: p) process == p.
	self assert: (profiler getThreadFrom: p) id = 1.
	
! !

!LoomProfilerTest methodsFor: 'tests' stamp: 'AlexandreBergel 9/5/2012 11:11'!
testGetLoomThreads
	| p ps |
	p := [ ] fork.

	ps := profiler getLoomThreadsFrom: (Array with: p).

	self assert: ps size = 1.
	self assert: ps first process == p.
	self assert: ps first id = 1.
! !

!LoomProfilerTest methodsFor: 'tests' stamp: 'AlexandreBergel 9/5/2012 11:40'!
testIsProfiling

	self deny: profiler isProfiling.
	profiler startProfiling.
	self assert: profiler isProfiling.
	
	profiler stopProfiling.
	self deny: profiler isProfiling.
	
! !

!LoomProfilerTest methodsFor: 'tests' stamp: 'AlexandreBergel 9/6/2012 09:03'!
testPrintingThreads

	self shouldnt: [ profiler allCurrentThreads do: [ :t | t printString ] ] raise: Error! !

!LoomProfilerTest methodsFor: 'tests' stamp: 'AlexandreBergel 9/6/2012 08:42'!
testPriority
	| p ps thread |
	p := [ ] fork.
	profiler profile: (Array with: p).
	thread := profiler getThreadFrom: p.

	self assert: thread priority = p priority! !

!LoomProfilerTest methodsFor: 'thread state' stamp: 'AlexandreBergel 9/6/2012 08:45'!
testThreadState

	| loomThread |
	self scenario1.
	profiler profile: (Array with: thread1).
	loomThread := profiler getThreadFrom: thread1. 
	
	self deny: thread1 isTerminated.
	self deny: loomThread isTerminated.
	
	self waitForScenarioEnd.
	
	self assert: thread1 isTerminated.
	self assert: loomThread isTerminated.
! !

!LoomProfilerTest methodsFor: 'util' stamp: 'AlexandreBergel 9/5/2012 09:55'!
wait
	(Delay forMilliseconds: 2) wait! !

!LoomProfilerTest methodsFor: 'util' stamp: 'AlexandreBergel 9/5/2012 09:55'!
wait2
	(Delay forMilliseconds: 4) wait! !

!LoomProfilerTest methodsFor: 'util' stamp: 'AlexandreBergel 9/5/2012 14:25'!
wait3
	(Delay forMilliseconds: 8) wait! !

!LoomProfilerTest methodsFor: 'util' stamp: 'AlexandreBergel 9/5/2012 10:12'!
waitForScenarioEnd
	[ self isScenarioOver not ] whileTrue: [ self wait ]! !

Object subclass: #LoomObject
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Loom'!

!LoomObject class methodsFor: 'instance creation' stamp: 'AlexandreBergel 9/6/2012 09:34'!
new
	^ self basicNew initialize! !

!LoomObject methodsFor: 'initialize-release' stamp: 'AlexandreBergel 9/6/2012 09:34'!
initialize! !

LoomObject subclass: #LoomProfiler
	instanceVariableNames: 'profiledThreads beginingProfilingTime profilingTime profilingProcess numberOfSamplings announcer'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Loom'!

!LoomProfiler methodsFor: 'private' stamp: 'AlexandreBergel 9/6/2012 09:37'!
allCurrentProcesses 
	"Return a weak array of processes"
	
	^" WeakSet withAll:"
		(Process allInstances
				reject: [:each | each isTerminated]).! !

!LoomProfiler methodsFor: 'private' stamp: 'AlexandreBergel 9/6/2012 09:02'!
allCurrentThreads
	"Return a weak array of loom threads"

	^ self getLoomThreadsFrom: self allCurrentProcesses! !

!LoomProfiler methodsFor: 'events' stamp: 'AlexandreBergel 9/11/2012 22:04'!
announce: anEvent

	self announcer announce: anEvent! !

!LoomProfiler methodsFor: 'accessing' stamp: 'AlexandreBergel 9/11/2012 21:44'!
announcer
	^ announcer ifNil: [ announcer := Announcer new ]! !

!LoomProfiler methodsFor: 'accessing computed' stamp: 'AlexandreBergel 9/5/2012 11:58'!
consumptionTimeOf: aSmalltalkProcess
	^ ((self getThreadFrom: aSmalltalkProcess) tick / numberOfSamplings) asFloat! !

!LoomProfiler methodsFor: 'private' stamp: 'AlexandreBergel 9/6/2012 10:16'!
createProfilingThread
	profilingProcess := [ [ self isProfiling ] whileTrue: [ 
							self sampleExecution.
							(Delay forMilliseconds: 1) wait.
						 ] ] forkAt: 72.
! !

!LoomProfiler methodsFor: 'private' stamp: 'AlexandreBergel 9/5/2012 19:29'!
getLoomThreadsFrom: processes
	"Return a collection of LoomThreads from a list of Smalltalk processes"

	| id |
	id := 0.
	^ processes collect: [ :p |
		id := id + 1.
		LoomThread new
			process: p;
			id: id;
			profiler: self;
			yourself ]! !

!LoomProfiler methodsFor: 'public' stamp: 'AlexandreBergel 9/5/2012 11:11'!
getThreadFrom: aSmalltalkProcess
	"Return the loom thread associated to the smalltalk process"
	
	^ profiledThreads detect: [ :t | t process == aSmalltalkProcess ]! !

!LoomProfiler methodsFor: 'public' stamp: 'AlexandreBergel 9/5/2012 14:54'!
getThreadFrom: aSmalltalkProcess ifAbsent: aBlock
	"Return the loom thread associated to the smalltalk process"
	profiledThreads isNil ifTrue: [ aBlock value ].
	^ profiledThreads detect: [ :t | t process == aSmalltalkProcess ] ifNone: aBlock! !

!LoomProfiler methodsFor: 'initialize-release' stamp: 'AlexandreBergel 9/6/2012 09:04'!
initialize
	super initialize.
	profilingTime := 0.
	beginingProfilingTime := 0.
	numberOfSamplings := 1! !

!LoomProfiler methodsFor: 'testing' stamp: 'AlexandreBergel 9/5/2012 11:41'!
isProfiling
	"Are we profiling?"
	
	^ beginingProfilingTime ~= 0! !

!LoomProfiler methodsFor: 'accessing' stamp: 'AlexandreBergel 9/5/2012 19:32'!
numberOfSamplings
	^ numberOfSamplings! !

!LoomProfiler methodsFor: 'events' stamp: 'AlexandreBergel 9/11/2012 22:03'!
on: anEvent do: aBlock
	"Register the callback one arg aBlock for the event anEvent, a subclass of LoomEvent"
	
	self announcer on: anEvent do: aBlock! !

!LoomProfiler methodsFor: 'public' stamp: 'AlexandreBergel 9/11/2012 21:38'!
profile: processCollection
	"profile a collection of Smalltalk processes"
	
	profiledThreads := self getLoomThreadsFrom: processCollection.
	self startProfiling! !

!LoomProfiler methodsFor: 'public' stamp: 'AlexandreBergel 9/11/2012 21:38'!
profileAllProcesses
	"profile all the active processes"
	
	self profile: self allCurrentProcesses! !

!LoomProfiler methodsFor: 'accessing' stamp: 'AlexandreBergel 9/5/2012 10:36'!
profilingTime
	^ profilingTime! !

!LoomProfiler methodsFor: 'private' stamp: 'AlexandreBergel 9/11/2012 22:19'!
sampleExecution 
	"Simply increases the tick of the loom thread next to be executed"

	| loomThread nextProcessToRun |
	nextProcessToRun := Processor nextReadyProcess.
	self announce: LoomSamplingEvent.
	
	nextProcessToRun isNil ifTrue: [ ^ self ].
	loomThread := self getThreadFrom: nextProcessToRun ifAbsent: [ ^ self ].
	loomThread increaseTick.
	numberOfSamplings := numberOfSamplings + 1.

	self announce: LoomSampledEvent! !

!LoomProfiler methodsFor: 'public' stamp: 'AlexandreBergel 9/11/2012 21:42'!
startProfiling
	"Start the monitoring"
	
	self startTimer.
	self createProfilingThread! !

!LoomProfiler methodsFor: 'private' stamp: 'AlexandreBergel 9/6/2012 09:50'!
startTimer
	beginingProfilingTime := LoomTime currentTimeAsMilliseconds! !

!LoomProfiler methodsFor: 'public' stamp: 'AlexandreBergel 9/11/2012 21:42'!
stopProfiling
	"Stop the monitoring"
	
	profilingTime := LoomTime currentTimeAsMilliseconds - beginingProfilingTime.
	beginingProfilingTime := 0! !

!LoomProfiler methodsFor: 'public' stamp: 'AlexandreBergel 9/11/2012 21:43'!
threadsDo: aBlock
	"Evaluate a block for each loom thread"
	
	profiledThreads do: aBlock! !

LoomObject subclass: #LoomThread
	instanceVariableNames: 'process id tick profiler'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Loom'!

!LoomThread methodsFor: 'accessing computed' stamp: 'AlexandreBergel 9/5/2012 19:40'!
consumptionTime
	^ (((tick / profiler numberOfSamplings) * 100) asInteger / 100) asFloat! !

!LoomThread methodsFor: 'accessing' stamp: 'AlexandreBergel 9/5/2012 10:58'!
id
	"Return the id of the thread. If 0, then it means that none has been set"

	^ id! !

!LoomThread methodsFor: 'accessing' stamp: 'AlexandreBergel 9/5/2012 10:58'!
id: aNumber
	"Set the id number"
	
	id := aNumber! !

!LoomThread methodsFor: 'public' stamp: 'AlexandreBergel 9/5/2012 11:50'!
increaseTick
	tick := tick + 1! !

!LoomThread methodsFor: 'initialize' stamp: 'AlexandreBergel 9/5/2012 11:50'!
initialize
	super initialize.
	id := 0.
	tick := 0! !

!LoomThread methodsFor: 'testing' stamp: 'AlexandreBergel 9/6/2012 08:45'!
isSuspended
	^ process isSuspended! !

!LoomThread methodsFor: 'testing' stamp: 'AlexandreBergel 9/6/2012 08:45'!
isTerminated
	"Return true if the process is terminated"
	
	^ process isTerminated! !

!LoomThread methodsFor: 'printing' stamp: 'AlexandreBergel 9/6/2012 09:39'!
printOn: stream
	stream nextPutAll: '['.
	stream nextPutAll: (self consumptionTime * 100) asInteger printString.
	stream nextPutAll: '%] ('.
	stream nextPutAll: self priority printString.
	stream nextPutAll: ')'.
	
	self isTerminated 
		ifTrue: [ stream nextPutAll: ' T '].
	
	(process suspendedContext class ==  MethodContext)
		ifTrue: [ stream nextPutAll: process suspendedContext home printString ]
		ifFalse: [ stream nextPutAll: process suspendedContext printString.]
	! !

!LoomThread methodsFor: 'accessing computed' stamp: 'AlexandreBergel 9/6/2012 08:42'!
priority
	"Answer the priority of the receiver."

	^ process priority! !

!LoomThread methodsFor: 'accessing' stamp: 'AlexandreBergel 9/5/2012 10:58'!
process
	^ process! !

!LoomThread methodsFor: 'accessing' stamp: 'AlexandreBergel 9/5/2012 10:58'!
process: anObject
	process := anObject! !

!LoomThread methodsFor: 'accessing' stamp: 'AlexandreBergel 9/5/2012 19:29'!
profiler: aProfiler
	profiler := aProfiler! !

!LoomThread methodsFor: 'public' stamp: 'AlexandreBergel 9/6/2012 09:15'!
resetTick
	"Invoked after each sampling"
	tick := 0! !

!LoomThread methodsFor: 'accessing' stamp: 'AlexandreBergel 9/5/2012 11:51'!
tick
	^ tick! !
